import java.lang.Math;
import java.util.*;


/**
 *
 * @author Alex Motchoulski
 */
public class DistributionSimulation {

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
	
    //parameter for max population
    int pop_max = 100;
    
    //the distribution that is a result, players will use this to gauge cost of injustice 
    int dist = 0;
    
    //global parameter that keeps track of previous steps' distribution 
    int prev_dist = 0;
    
    //maximum range of toleration that any given agent can have 
    //as of now, toleration is used to calculate whether an agent becomes more egoistical at the end of the round 
    int max_tol = 5;
    
    //parameter that sets the number of steps that the simulation will run 
    int step_max = 21;
    
    // the range of distribution positions
    int dist_range = 24;
    
    // parameter that sets the weight of individuals signals in the power case, if an agent's power_dist is the same as the current dist
    //then their vote is weighed three times, if it is one off, it is weighed 2 times, and regular weight for all others 
    int power_range = 4;
    
    //boolean parameter for tracking if the loop is in the first run 
    boolean first = true;
    
    
    //----------------------------CASE PARAMETERS----------------------------------------------------------------------
    
	//thed efault version of the model, power is not in play with the voting rule, no unique kinds of agents (strategic etc.)  
    // T= no power dimension, F = power via extra votes
	boolean def_case = false;
    
    // sets whether there is lag in the shift calculuation, 
	// T = shift will be incomplete, i.e. divided by some value (either population or opposition)
    boolean social_inertia = false; 
    
    // sets whether social intertia is weigthed by the total population, or only the opposition
    // T = divide by max population, F =  divide by opposition, easier shift compared to max case 
    boolean pop_inertia = false;
    
   // change these into probability calculuations when messing with population distribution of the various characteristics
    double strategic_chance = .1;
    boolean adaptive_chance = false;
    
    
    boolean contract_case = false;
    
    boolean borda_case = false;
    
    boolean plurality_case = false;
    
    //----------------------------------------------------------------------------------------------------------------

    
    ArrayList<Integer> signals = new ArrayList<Integer>();
    ArrayList<Agent> pop = new ArrayList<Agent>();
    
    int strat =0;
	//initial populating of the Agent array, each agent in the initial population is random, change code to make new kinds of tests, e.g. set all probability of strategic agents to 0
    //parameters for agents are: preferred distribution, ego_pref, tolerance range, strategic, adaptive
    for(int i = 0; i <= pop_max; i++){
        Agent w = new Agent((int)(Math.random() * dist_range),(int)(Math.random() * dist_range) , (int)(Math.random() * max_tol), Math.random() < strategic_chance, adaptive_chance);
        pop.add(w);
        
        if(w.is_strategic()) {
        	 strat++;
        }
       
      }
    System.out.println("number strat = " + strat);
    
    //the simulation loop, runs for some number of steps, set by  step_max-- each of the loops corresponds to a different type of simulation 
    for(int i = 0; i < step_max; i++){
    	
        //start with a first round check here, if yest do a random distribution, else send it to the main computation
        if(first){
            dist = (int) (Math.random() * dist_range);
            first = false;
            System.out.println("The distribution for round " + i + " is " + dist);
            for(int j = 0; j < pop.size(); j++){
        			if(pop.get(j).is_strategic() && borda_case) {
        				pop.get(j).instantiate_borda_strat(dist_range);
        			}
            }
        }
        // signal collection for the default case: each signal is given equal weight, and the median signal is selected 
        else if(def_case){
        //signal collection for the current population, will do non-random distribution after first 
            for(int j = 0; j < pop.size(); j++){
                  //check for if the agent is playing their first round, if not hindsight and strategic elements come into play
                  if(!pop.get(j).is_first_round() && pop.get(j).is_strategic()){
                     //check if agent will send a strategic signal
                	  	// in the plurality vote, the agent will vote for what appears to be the front runner
                	  	
                	  
                	  	if(plurality_case){
                	  		int lead = 0;
                  	  	int lead_appearance = 0;		
                        	  	for(int k = pop.get(j).get_pref_dist()-pop.get(j).get_tolerance(); k <= pop.get(j).get_pref_dist() + pop.get(j).get_tolerance(); k++ ) {
                        	  		int temp_appear = 0;
                        	  		for(int h = 0; h < signals.size(); h++) {
                        	  			if(signals.get(h) == k) {
                        	  				temp_appear++;
                        	  			}
                        	  		}
                        	  		
                        	  		if(temp_appear > lead_appearance) {
                        	  			lead = k;
                        	  			lead_appearance = temp_appear;
                        	  		}
                        	  	}
                        	  	if(lead_appearance > 0) {
                        	  		signals.add(lead);
                        	  	}
                        	  	else {
                        	  		signals.add(pop.get(j).get_pref_dist());
                        	  	}
                	  	}
                	  	else {
                	  		signals.add(pop.get(j).strategic_median_signal(prev_dist));
                	  	}
                   }
                  
                    // if agents are in the first round, they will always signal honestly
                    else{
                    		signals.add(pop.get(j).get_pref_dist());
                    }      
            }
            
             Collections.sort(signals);
             
             
             // *****VOTING RULES*****
             // 1. contract case, 2. borda count, 3. simple plurality, 4. median 
             
             int proposed_dist = 0;
             
             //contract case: selects the distribution that will minimize egoistic shift among voters, e.g. looks for terms that are exceptable to largest number of agents
             if(contract_case){
            	 	int num_shift = pop_max;
            	 	int prop_shift = 0;
            	 	for(int j = 0; j < dist_range; j ++){
            	 		prop_shift = 0;
            	 		for(int k =0; k< pop.size(); k++) {
            	 			//checks for strategic agents, in the contract case a strategic agent will be misrepresent their tolerance, showing all but their optimum as intolerable
            	 			if(pop.get(k).is_strategic() && j != pop.get(k).get_pref_dist()) {
            	 				prop_shift++;
            	 			}
            	 			else if(pop.get(k).will_ego_shift(j)){
            	 				prop_shift++;
            	 			}
            	 		}
            	 		if(prop_shift < num_shift) {
            	 			num_shift = prop_shift;
            	 			proposed_dist = j;
            	 		}
            	 		else if(prop_shift == num_shift){
            	 			if(Math.random() > .5) {
            	 				num_shift = prop_shift;
                	 			proposed_dist = j;
            	 			}
            	 		}	
            	 	}
            	 	System.out.println("Contract selection");
             }
             // borda count vote: highest option gets dist_range votes, preferences determined by linear distance from ideal 
             else if(borda_case) {
            	 int score = 0;
            	    ArrayList<Integer> votes = new ArrayList<Integer>();
            	 	for(int j = 0; j< pop.size(); j++) {
            	 		if(pop.get(j).is_strategic() && !pop.get(j).is_first_round()) {
            	 			ArrayList<Integer> borda_strat = pop.get(j).get_borda_votes();
            	 			for(int k = 0; k < borda_strat.size(); k++) {
            	 				score = dist_range - borda_strat.indexOf(k);
            	 				for(int h = 0; h < score; h++ ) {
            	 					votes.add(k);
            	 				}
            	 			}
            	 		}
            	 		
            	 		else {
            	 			for(int k = 0; k< dist_range; k++) {
                	 			score = dist_range - Math.abs(pop.get(j).get_pref_dist() - k);
                	 			for(int h = 0 ; h < score; h++) {
                	 				votes.add(k);
                	 			}
                	 		}
            	 		}
            	 		
            	 	}
            	 	proposed_dist = mode(votes, dist_range);
            	 	
            	 	System.out.println("Borda selection");
             }
             // plurality: takes the mode signal
             else if(plurality_case) {
            	 	proposed_dist = mode(signals, dist_range);
            	 	System.out.println("Plurality selection");
             }
             // median case: treated as default: takes median signal 
             else {
            	 proposed_dist = signals.get(pop_max/2);
            	 System.out.println("Median selection");
             }
                          
 
             //sticky case 
             if(social_inertia) {
            	 int support_direction = 0;
                 for(int k = 0; k < signals.size(); k++){
                     if(proposed_dist > dist){
                       if(signals.get(k) >= proposed_dist){
                           support_direction++;
                       }   
                     }
                     else{
                        if(signals.get(k) <= proposed_dist){
                           support_direction++; 
                        } 
                     }
                 }
                 int opposition =  pop_max - support_direction;
                 
                 //catching divide by zero errors 
                 if(opposition == 0){ opposition = 1;}
                 
                 // pop_inertia reduces shift by a higher degree, models stickiness of distribution for whole population
                 // not pop_inertia models case where those in support of the move reduce the stickiness 
                 if(pop_inertia) {
                	 	int temp_prop = dist + (int)(((proposed_dist-dist) * support_direction)/(pop_max));
                	 
                	 	//catches for overshooting 
                	 	if(proposed_dist-dist < 0) {
                	 		if(temp_prop < proposed_dist) {
                	 			dist = proposed_dist;
                		 	}
                	 	}
                	 	else if (proposed_dist-dist > 0) {
                	 		if(temp_prop > proposed_dist) {
                	 			dist = proposed_dist;
                	 		}
                	 	}
                	 
                	 	if(dist > dist_range) {
                	 		dist = dist_range;
                	 	}
                	 	if(dist < 0) {
                	 		dist = 0;
                	 	}
                 }
                 else {
                	 int temp_prop = dist + (int)(((proposed_dist-dist) * support_direction)/(opposition));
              
                	 //catches for overshooting 
                	 if(proposed_dist-dist < 0) {
                		 if(temp_prop < proposed_dist) {
                			 dist = proposed_dist;
                		 }
                	 }
                	 else if (proposed_dist-dist > 0) {
                		 if(temp_prop > proposed_dist) {
                			 dist = proposed_dist;
                		 }
                	 }
                	 
                	 	if(dist > dist_range) {
             	 		dist = dist_range;
             	 	}
             	 	if(dist < 0) {
             	 		dist = 0;
             	 	}
                 }
                 
             }
             //the non-sticky case, shift moves completely to the proposed 
             else { 
            	 	dist = proposed_dist;
             }
             
             System.out.println("The distribution for round " + i + " is " + dist);  
        }
        
        
  //POWER CASE: the votes of agents are given increased weight if the distirbution is closer to their 'power distribution,'
        else{
        	 int proposed_dist = 0;
        	 
        	 for(int j = 0; j < pop.size(); j++) {
        		 
        		 if(pop.get(j).is_strategic()){
                     //check if agent will send a strategic signal
        			 
                	  	// in the plurality vote, the agent will vote for what appears to be the front runner
                	  	if(plurality_case){
                	  		int lead = 0;
                  	  	int lead_appearance = 0;		
                        	  	for(int k = pop.get(j).get_pref_dist()-pop.get(j).get_tolerance(); k <= pop.get(j).get_pref_dist() + pop.get(j).get_tolerance(); k++ ) {
                        	  		int temp_appear = 0;
                        	  		for(int h = 0; h < signals.size(); h++) {
                        	  			if(signals.get(h) == k) {
                        	  				temp_appear++;
                        	  			}
                        	  		}
                        	  		
                        	  		if(temp_appear > lead_appearance) {
                        	  			lead = k;
                        	  			lead_appearance = temp_appear;
                        	  		}
                        	  	}
                        	  	if(lead_appearance > 0) {
                        	  		if(Math.abs(pop.get(j).get_power_dist() - dist) < power_range) {
                           			 for(int k = 0; k < Math.abs(pop.get(j).get_power_dist() - dist); k++){
                           				signals.add(lead);
                           			 }
                           		 }
                        	  		else {
                        	  			signals.add(lead);
                        	  		}
                        	  	}
                        	  	else {
                        	  		if(Math.abs(pop.get(j).get_power_dist() - dist) < power_range) {
                              			 for(int k = 0; k < Math.abs(pop.get(j).get_power_dist() - dist); k++){
                              				signals.add(pop.get(j).get_pref_dist());
                              			 }
                              		 }
                        	  		else {
                        	  			signals.add(pop.get(j).get_pref_dist());
                        	  		}
                        	  	}
                	  	}
                	  	else { //median strategic signalling
                	  		if(Math.abs(pop.get(j).get_power_dist() - dist) < power_range) {
                     			 for(int k = 0; k < Math.abs(pop.get(j).get_power_dist() - dist); k++){
                     				signals.add(pop.get(j).strategic_median_signal(prev_dist));
                     			 }
                     		 }
                	  		else {
                	  			signals.add(pop.get(j).strategic_median_signal(prev_dist));
                	  		}
                	  	}
                   }
                  
                    // non-strategic power signalling
                    else{ 
                    	if(Math.abs(pop.get(j).get_power_dist() - dist) < power_range) {
                 			 for(int k = 0; k < Math.abs(pop.get(j).get_power_dist() - dist); k++){
                 				signals.add(pop.get(j).get_pref_dist());
                 			 }
                 		 }
           	  		else {
           	  			signals.add(pop.get(j).get_pref_dist());
           	  		}
                    }
        		 
        		 // checks if the agent has more 'power' in which case they signal more
        		 if(Math.abs(pop.get(j).get_power_dist() - dist) < power_range) {
        			 for(int k = 0; k < Math.abs(pop.get(j).get_power_dist() - dist); k++){
        				 signals.add(pop.get(j).get_pref_dist());
        			 }
        		 }
        		 else {
        			signals.add(pop.get(j).get_pref_dist());
        		 }
        	 }
        	 
        	 
        	 // Voting Rules, same as above, with changes to reflect power
        	 
        	 //contract power case: those in power win ties in acceptability-- their 'stubborness' carries more weight 
        	 if(contract_case){
         	 	int num_shift = pop_max;
         	 	int power_max = 0;
         	 	for(int j = 0; j < dist_range; j ++){
         	 		int prop_shift = 0;
         	 		int power_count = 0;
         	 		for(int k =0; k< pop.size(); k++) {
         	 			if(pop.get(k).is_strategic() && j != pop.get(k).get_pref_dist()) {
         	 				prop_shift++;
         	 			}
         	 			else if(pop.get(k).will_ego_shift(j)){
         	 				prop_shift++;
         	 			}
         	 			// tracks how much 'power' there is behind the curent proposed distribution    
         	 			if(Math.abs(pop.get(j).get_power_dist() - dist) < power_range) {
         	 				power_count+= Math.abs(pop.get(j).get_power_dist() - dist);
         	 				//System.out.println("This agent's power_dist = " + pop.get(j).get_power_dist() + " the agent has this much more power: " + (power_range - Math.abs(pop.get(j).get_power_dist() - dist)));
         
         	 				
         	 			}
         	 		}
         	 		if(prop_shift < num_shift) {
         	 			num_shift = prop_shift;
         	 			proposed_dist = j;
         	 		}
         	 		//the power additions
         	 		else if(prop_shift == num_shift){
         	 			if(power_count > power_max) {
         	 				num_shift = prop_shift;
             	 			proposed_dist = j;
             	 			power_max = power_count;
         	 			}
         	 			else if(power_count == power_max) {
         	 				if(Math.random() > .5) {
             	 				num_shift = prop_shift;
                 	 			proposed_dist = j;
             	 			}
         	 			}
         	 		}  
         	 		
         	 
         	 	}
         	 	System.out.println("Contract-power selection");
          }
        	 
          // borda power case: if in power get extra votes for highest option 
          else if(borda_case) {
         	 int score = 0;
         	    ArrayList<Integer> votes = new ArrayList<Integer>();
         	    
         	   for(int j = 0; j< pop.size(); j++) {
       	 		if(pop.get(j).is_strategic() && !pop.get(j).is_first_round()) {
       	 			ArrayList<Integer> borda_strat = pop.get(j).get_borda_votes();
       	 			for(int k = 0; k < borda_strat.size(); k++) {
       	 				score = dist_range - borda_strat.indexOf(k);
       	 				
       	 			//this conditional is an addition to the power case, most preferred option gets extra votes based on proximity to the agent's power distribution
       	 				if(score == dist_range && Math.abs(pop.get(j).get_power_dist() - dist) < (power_range)) {
       	 					score += Math.abs(pop.get(j).get_power_dist() - dist);	
       	 				}
       	 				for(int h = 0; h < score; h++ ) {
       	 					votes.add(k);
       	 				}
       	 			}
       	 		}
       	 		
       	 		else {
       	 			for(int k = 0; k< dist_range; k++) {
           	 			score = dist_range - Math.abs(pop.get(j).get_pref_dist() - k);
           	 			
           	 			if(score == dist_range && Math.abs(pop.get(j).get_power_dist() - dist) < (power_range)) {
           	 				score += Math.abs(pop.get(j).get_power_dist() - dist);
           	 			}
           	 			
           	 			for(int h = 0 ; h < score; h++) {
           	 				votes.add(k);
           	 			}
           	 		}
       	 		}
       	 		
         	   }
         	    
         	 	proposed_dist = mode(votes, dist_range);
         	 	System.out.println("Borda-power selection");
          }
          // plurality: takes the mode signal, since it is based on signals, the relevant power changes are already included
          else if(plurality_case) {
         	 	proposed_dist = mode(signals, dist_range);
         	 	System.out.println("Plurality-power selection");
          }
          // median case: treated as default: takes median signal, same as plurality case, already accounts for power changes in signal calculation
          else {
         	 proposed_dist = signals.get(pop_max/2);
         	System.out.println("Median-power selection");
          }
                       
        	  
       
        	  // the social inertia case, reduces the effectiveness of the movement based on either population or opposition
        	  if(social_inertia) {
        		  int support_direction = 0;
                  for(int k = 0; k < signals.size(); k++){
                      if(proposed_dist > dist){
                        if(signals.get(k) >= proposed_dist){
                            support_direction++;
                        }   
                      }
                      else{
                         if(signals.get(k) <= proposed_dist){
                            support_direction++; 
                         } 
                      }
                  }
                  int opposition =  pop_max - support_direction;
                  
                  //catching divide by zero errors 
                  if(opposition == 0){ opposition = 1;}
                  
                  
                  // the two different kinds of inertia, see documentaiton above in the default case 
                  if(pop_inertia) {
                   int temp_prop = dist + (int)(((proposed_dist-dist) * support_direction)/(pop_max));
                   
                   //catches for overshooting 
                   if(proposed_dist-dist < 0) {
                	   if(temp_prop < proposed_dist) {
              			 dist = proposed_dist;
                	   	}
                   }
                   else if (proposed_dist-dist > 0) {
              		 if(temp_prop > proposed_dist) {
              			 dist = proposed_dist;
              		 	}
              	 	}
              	 
                 	 if(dist > dist_range) {
                 		 dist = dist_range;
            	 		 }
                 	 if(dist < 0) {
                 		 dist = 0;
                 	 }
                  }
                  else {
                 	 dist = dist + (int)(((proposed_dist-dist) * support_direction)/(opposition));
                 	 if(dist > dist_range) {
                 		 dist = dist_range;
            	 			}
            	 			if(dist < 0) {
            	 				dist = 0;
            	 			}
                  }
        	  }
        	  else {
        		  dist = proposed_dist;
        	  }
        	  
        	  
        	  System.out.println("The distribution for round " + i + " was " + dist);
        	 
        }
     
      
      for(int j = 0; j < pop.size(); j++){
          
          
          //adaptive agents will move toward the resulting distibution if their benefit was positive
          if(pop.get(j).is_adaptive()){
              pop.get(j).adapt(dist, dist_range);
             // System.out.println("someone adapted");
          }
          
          //agents who find the current distribution intolerable will shift their preferred distribution towards their power distribution 
          // adaptive agents do not experience ego shifts 
          
          if(Math.abs(dist - pop.get(j).get_pref_dist()) > pop.get(j).get_tolerance() && !pop.get(j).is_adaptive() && !pop.get(j).is_first_round()) {
        	  	pop.get(j).ego_shift();
          }
          
        //tracks that the agent is no longer playing the first round, strategic agents will now play modified strategies 
          pop.get(j).set_round();
          
          if(pop.get(j).is_strategic() && borda_case) {
        	  	pop.get(j).borda_manip(dist_range, dist, mode(signals, dist_range));
          }
          
      }
           
     int total_egoists = 0;
     int ego_shifters = 0;
     
     for(int j = 0; j < pop.size(); j++) {
    	 	if(pop.get(j).is_egoistic()) {
    	 		total_egoists++;
    	 	}
    	 	if(pop.get(j).has_ego_shifted()) {
    	 		ego_shifters++;
    	 	}
     }
     
     System.out.println("The number of agents who experienced an ego shift is " + ego_shifters);
     System.out.println("The number of agents who are egoistic is " + total_egoists);
      
      prev_dist = dist;
      signals.clear();
      
      System.out.println(" ");
    }
    //end of simulation loop
  
  }
    
  
    public static int mode(ArrayList<Integer> set, int d_range) {
        int mode = 0;
        int max_count = 0;
        
        for(int i =0; i< d_range; i++) {
       	 	int element = i;
       	 	int count = 0;
       	 	
       	 	for(int j = 0; j < set.size(); j++) {
       	 		if(set.get(j) == element) {
       	 			count ++;
       	 		}
       	 	}
       	 	
       	 	if(count >max_count) {
       	 		mode = element;
       	 		max_count = count;
       	 	}
       	 	// random selection in case of ties, with equal chance to each option 
       	 	else if(count == max_count) {
       	 		if(Math.random() > 0.5) {
       	 			mode = element;
       	 		}
       	 	}  
      }
    	 return mode;

    }

}